{$X+} { Turn on extended syntax }
Program GBChrEd;

(**********************************************************************
*
* GBChrEd V01.00.00
* Programmed by David Dahl aka TechT
*
* PUBLIC DOMAIN
*
* A simple Game Boy 8x8 character set editor
*
* Compile using Borland Pascal 7 in *REAL MODE*
*
* Started : 10/29/96
* Finished: ??/??/??
*
* Notes:
*
*   This program won't work in protected mode 'cause I access a real
*   mode segment to get the IBM character table info.  I *could* use
*   DPMI calles to get around that, I know, but I don't have any docs
*   on DPMI programming.  Maybe later.
*
*   This program works well enough to edit character sets, but I still
*   haven't implemented all the options I'd like to implement.
*
*   I also should make the undo buffer smart so it only saves changes,
*   not a copy of the entire character buffer.  (Those little 4k slices
*   really add up in real mode.)
*
**********************************************************************)

Uses
  CRT, Strings;

Const
  Version = '01.00.00';

Const
  Key_UpArrow    = $48;
  Key_DownArrow  = $50;
  Key_LeftArrow  = $4B;
  Key_RightArrow = $4D;

  Key_Home       = $47;
  Key_End        = $4F;

  Key_PageUp     = $49;
  Key_PageDown   = $51;

  Key_Delete     = $53;

Const
  MaxUndo = 5;

Const
  CharSetFileId : PChar = 'Game Boy 8x8 character set image file.'+
                          #13+#10+#26;

{=[ TYPE DEFINITIONS ]================================================}
{-[ For Holding Gameboy Character Bit Patterns ]----------------------}
Type
  GBChar    = Array[ 0..15 ] of Byte;
  GBCArray  = Array[ 0..255 ] of GBChar;
  pGBCArray = ^GBCArray;

{-[ For Undo Buffer ]-------------------------------------------------}
Type
  UndoLink  = ^UndoRec;
  UndoRec   = Record
    Next     : UndoLink;
    Data     : GBCArray;
  End;

{-[ For Quick Gameboy Character Patterns to VGA Display ]-------------}
Type
  GB2VGAArray  = Array[ 0..$FF, 0..7 ] of Byte;
  pGB2VGAArray = ^GB2VGAArray;
  GB2VGApTbl   = Array[ 0..$FF ] of pGB2VGAArray;

{-[ For VGA Palette ]-------------------------------------------------}
Type
  RGBRec = Record
    Red   : Byte;
    Green : Byte;
    Blue  : Byte;
  End;
  GBColorArray = Array[0..3] of RGBRec;

{-[ For VGA Screen Access ]-------------------------------------------}
Type
  VGAPtr   = ^VGAArray;
  VGAArray = Array[ 0..199, 0..319 ] of Byte;

{-[ Misc ]------------------------------------------------------------}
Type
  ColorRec = Record
    B : Byte;
    F : Byte;
  End;

  BIOSCArray  = Array[ 0..255, 0..7 ] of Byte;
  pBIOSCArray = ^BIOSCArray;

Type
  eCharWinState = ( W32x8, W16x16, W32x8Interlaced, W16x16Interlaced );

{-[ For Cursors ]-----------------------------------------------------}
Type
  CursorRec = Record
    Index : Word;
    X, Y  : Integer;
    Color : Byte;
  End;

  PickBkgRec = Record
    X, Y : Integer;
    Data : Array [ 0..9, 0..9 ] of Byte;
  End;

{=[ GLOBAL VARIABLES ]================================================}
Const
  MainCounter : Word = 0;

  WorkChanged : Boolean = FALSE;

Var
  CurrentFile : String;

  CharBuffer  : pGBCArray;
  UndoBuffer  : UndoLink;
  CopyChar    : GBChar;

  Done        : Boolean;
  VGA         : VGAPtr;
  GB2VGATbl   : GB2VGApTbl;

  BIOS8X8     : pBIOSCArray;

  PickBkgBuff : PickBkgRec;

Const
  Cursor1    : CursorRec = ( Index : 0;
                             X     : 0;
                             Y     : 0;
                             Color : Red );
  Cursor2    : CursorRec = ( Index : 0;
                             X     : 0;
                             Y     : 0;
                             Color : Blue );
Const
  CharWinState : eCharWinState = W32x8;

Const
  BackgroundColor : ColorRec = ( B : DarkGray;   F : LightGray );
  StatusColor     : ColorRec = ( B : LightGray;  F : Black     );
  MenuColor       : ColorRec = ( B : LightBlue;  F : White     );

Const
  GBColors : GBColorArray = ( ( Red : 35; Green : 42; Blue :  0 ),
                              ( Red : 25; Green : 35; Blue :  6 ),
                              ( Red : 16; Green : 30; Blue : 12 ),
                              ( Red : 11; Green : 22; Blue : 20 ) );

{-[ PROCEDURES AND FUNCTIONS ]========================================}
function FileExists(FileName: String): Boolean;
{ Boolean function that returns True if the file exists;otherwise,
  it returns False. Closes the file if it exists. }
var
  F: file;
begin
  {$I-}
  Assign(F, FileName);
  FileMode := 0;  { Set file access to read only }
  Reset(F);
  Close(F);
  {$I+}
  FileExists := (IOResult = 0) and (FileName <> '');
end;  { FileExists }

{-[ Draw A Box ]------------------------------------------------------}
Procedure Box( X,     Y,
               XSize, YSize : Integer;
               Color        : Byte );
Var YCount : Integer;
Begin
  FillChar( VGA^[ Y, X ], XSize, Color );
  FillChar( VGA^[ Y + YSize - 1, X ], XSize, Color );
  For YCount := 1 to ( YSize - 2 ) do
  Begin
    VGA^[ YCount + Y, X ] := Color;
    VGA^[ YCount + Y, X + XSize - 1 ] := Color;
  End;
End;

{-[ Draw A Solid Bar ]------------------------------------------------}
Procedure Bar( X,     Y,
               XSize, YSize : Integer;
               Color        : Byte );
Var YCount : Integer;
Begin
  For YCount := 0 to ( YSize - 1 ) do
    FillChar( VGA^[ Y + YCount , X ], XSize, Color );
End;
{-[ Draw Cursors ]----------------------------------------------------}

Procedure DrawZoomCursor ( CursorOn : Boolean );
Var X, Y : Integer;
Begin
  X := ( 8 * 18 ) + ( Cursor1.X * 4 );
  Y := ( 8 * 18 ) + ( Cursor1.Y * 4 );
  If CursorOn
  Then
    Box( X, Y, 5, 5, Cursor1.Color )
  Else
    Box( X, Y, 5, 5, StatusColor.B );
End;

Procedure GetPickBkg(     X, Y         : Integer;
                      Var Buffer       : PickBkgRec );
Var CY : Integer;
Begin
  Buffer.X := X;
  Buffer.Y := Y;

  For CY := 0 to 9 do
    Move( VGA^[ Y + CY, X ], Buffer.Data[ CY, 0 ], 10 );
End;

Procedure PutPickBkg( Var Buffer : PickBkgRec );
Var CY : Integer;
Begin
  For CY := 0 to 9 do
    Move( Buffer.Data[ CY, 0 ], VGA^[ Buffer.Y + CY, Buffer.X ], 10 );
End;

Procedure DrawPickCursor ( CursorOn : Boolean );
Var
  X, Y      : Integer;
Const
  FirstCall : Boolean = TRUE;
Begin
  X :=  4 * 8 - 1 + ( ( Cursor1.Index AND 31 ) * 8 );
  Y :=  11 * 4 - 1 + ( ( ( Cursor1.Index SHR 5 ) AND 7 ) * 8 );

  If CursorOn
  Then
  Begin
    If FirstCall
    Then
      FirstCall := FALSE
    Else
      PutPickBkg( PickBkgBuff );

    GetPickBkg( X, Y, PickBkgBuff );
    Box( X, Y, 10, 10, Cursor1.Color )
  End
  Else
    If Not( FirstCall )
    Then
    Begin
      PutPickBkg( PickBkgBuff );
      FirstCall := TRUE;
    End;
End;

{-[ Build GameBoy Char to VGA Translation Table ]---------------------}
Procedure BuildGB2VGATbl;
Var
  Tbl   : Byte;
  Value : Byte;
  Index : Byte;
  Q     : Byte;
Begin
  For Tbl := 0 to $FF do
  Begin
    New( GB2VGATbl[ Tbl ] );

    For Value := 0 to $FF do
      For Index := 0 to 7 do
      Begin
        Q := ( ( ( Tbl SHR Index ) AND 1 ) * 2 ) OR
             ( ( Value SHR Index ) AND 1 );
        GB2VGATbl[ Tbl ]^[ Value ][ 7 - Index ] := 16 + Q;
      End;
  End;
End;

{-[ Draw Text @ Specified (X,Y) and color ]---------------------------}
Procedure WriteTextXY ( X, Y  : Integer;
                        S     : String;
                        Color : Byte );
Var
  XCount : Integer;
  YCount : Integer;
  L      : Integer;
  Index  : Integer;
  Value  : Byte;
Begin
  L := Length(S);
  If L > 0
  Then
    For Index := 1 to L do
      For YCount := 0 to 7 do
      Begin
        Value := BIOS8X8^[ Ord( S[Index] ), YCount ];
        For XCount := 0 to 7 do
          If( ( ( Value SHL XCount ) AND 128 ) <> 0 )
          Then
            VGA^[ Y + YCount ,
                  X + XCount + ( ( Index - 1 ) * 8 ) ] :=
              Color;
      End;
End;

{-[ Update Entire Zoom Line ]-----------------------------------------}
Procedure DrawZoomAll;
Var
  CharNum     : Integer;
  Color       : Byte;
  XPos, YPos  : Integer;
  CX, CY, CY2 : Integer;

  ZoomChar    : GBChar;
Begin
  For CharNum := 0 to 3 do
  Begin
    ZoomChar := CharBuffer^[ ( CharNum + Cursor1.Index - 4 ) AND 255 ];
    XPos := 2 + ( CharNum * 4 * 8 ) + ( CharNum * 3 );
    For CY := 0 to 7 do
      For CX := 0 to 7 do
      Begin
        Color := 16 +
          ( ( ( ZoomChar[ CY * 2 ] SHR ( 7 - CX ) ) AND 1 ) OR
            ( ( ZoomChar[ CY * 2 + 1 ] SHR ( 7 - CX ) ) AND 1 ) SHL 1 );
        For CY2 := 0 to 2 do
          FillChar( VGA^[ 8 * 18 + 1 + CY2 + ( CY * 4 ),
                          1 + XPos + ( CX * 4 ) ],
                    3, Color );
      End;
  End;

  CharNum := 4;
  Begin
    ZoomChar := CharBuffer^[ ( CharNum + Cursor1.Index - 4 ) AND 255 ];
    XPos := ( CharNum * 4 * 8 ) + ( CharNum * 4 );
    For CY := 0 to 7 do
      For CX := 0 to 7 do
      Begin
        Color := 16 +
          ( ( ( ZoomChar[ CY * 2 ] SHR ( 7 - CX ) ) AND 1 ) OR
            ( ( ZoomChar[ CY * 2 + 1 ] SHR ( 7 - CX ) ) AND 1 ) SHL 1 );
        For CY2 := 0 to 2 do
          FillChar( VGA^[ 8 * 18 + 1 + CY2 + ( CY * 4 ),
                          1 + XPos + ( CX * 4 ) ],
                    3, Color );
      End;
  End;

  For CharNum := 5 to 8 do
  Begin
    ZoomChar := CharBuffer^[ ( CharNum + Cursor1.Index - 4 ) AND 255 ];
    XPos := ( CharNum * 4 * 8 ) + ( CharNum * 3 ) - ( 8 * 3 ) - 1 ;
    For CY := 0 to 7 do
      For CX := 0 to 7 do
      Begin
        Color := 16 +
          ( ( ( ZoomChar[ CY * 2 ] SHR ( 7 - CX ) ) AND 1 ) OR
            ( ( ZoomChar[ CY * 2 + 1 ] SHR ( 7 - CX ) ) AND 1 ) SHL 1 );
        For CY2 := 0 to 2 do
          FillChar( VGA^[ 8 * 18 + 1 + CY2 + ( CY * 4 ),
                          32 + XPos + ( CX * 4 ) ],
                    3, Color );
      End;
  End;
End;

{-[ Update Character In Active Zoom Box ]-----------------------------}
Procedure UpdateZoomActive( Index : Word );
Var
  ZoomChar    : GBChar;
  XPos        : Integer;
  CX, CY, CY2 : Integer;
  Color       : Byte;
Const
  CharNum = 4;
Begin
    ZoomChar := CharBuffer^[ Index  AND 255 ];
    XPos := ( CharNum * 4 * 8 ) + ( CharNum * 4 );
    For CY := 0 to 7 do
      For CX := 0 to 7 do
      Begin
        Color := 16 +
          ( ( ( ZoomChar[ CY * 2 ] SHR ( 7 - CX ) ) AND 1 ) OR
            ( ( ZoomChar[ CY * 2 + 1 ] SHR ( 7 - CX ) ) AND 1 ) SHL 1 );
        For CY2 := 0 to 2 do
          FillChar( VGA^[ 8 * 18 + 1 + CY2 + ( CY * 4 ),
                          1 + XPos + ( CX * 4 ) ],
                    3, Color );
      End;
End;

{-[ Draw The Background ]---------------------------------------------}
Procedure SetUpBackground;
Var CX, CY, CY2 : Integer;
    CharNum     : Integer;
    XPos, YPos  : Integer;
    Color       : Byte;
Begin
  FillChar( VGA^, 320 * 200, BackgroundColor.B );
  FillChar( VGA^, 320 * 8, StatusColor.B );
  FillChar( VGA^[ 8 * 18,0 ],
            320 * ( ( 8 * 4 ) + 1 ), StatusColor.B );
  Box( ( 8 * 16 ) + ( 4 * 4 ) - 1, 8 * 18 - 1,
       4 * 8 + 3 , 4 * 8 + 3, StatusColor.F );

  Box( 2, 8 * 14, 11, 7, StatusColor.F );
  For YPos := 1 to 5 do
    FillChar( VGA^[ 8 * 14 + YPos, 3 ], 9, 16 );
  WriteTextXY ( 8 * 2, 8 * 14,'Color 1', BackgroundColor.F );

  Box( 2, 8 * 15, 11, 7, StatusColor.F );
  For YPos := 1 to 5 do
    FillChar( VGA^[ 8 * 15 + YPos, 3 ], 9, 17 );
  WriteTextXY ( 8 * 2, 8 * 15,'Color 2', BackgroundColor.F );

  Box( 2, 8 * 16, 11, 7, StatusColor.F );
  For YPos := 1 to 5 do
    FillChar( VGA^[ 8 * 16 + YPos, 3 ], 9, 18 );
  WriteTextXY ( 8 * 2, 8 * 16,'Color 3', BackgroundColor.F );

  Box( 2, 8 * 17, 11, 7, StatusColor.F );
  For YPos := 1 to 5 do
    FillChar( VGA^[ 8 * 17 + YPos, 3 ], 9, 19 );
  WriteTextXY ( 8 * 2, 8 * 17,'Color 4', BackgroundColor.F );

  DrawZoomAll;
End;

{-[ Update Text On The Status Bar ]-----------------------------------}
Procedure DisplayStatus( S : String );
Var L       : Word;
    X, Y    : Word;
    Value   : Byte;
    Counter : Word;
    XPos    : Word;
Begin
  FillChar( VGA^, 320 * 8, StatusColor.B );
  L := Length(S);
  If L > 40
  Then
    L := 40;

  XPos := 160 - (L * 4);
  For Counter := 1 to L do
  Begin
    For Y := 0 to 7 do
    Begin
      { protected mode will barf on the next line }
      Value := BIOS8X8^[ Ord( S[Counter] ), Y ];
      For X := 0 to 7 do
        If( ( ( Value SHL X ) AND 128 ) <> 0 )
        Then
          VGA^[ Y, XPos + X ] := StatusColor.F;
    End;
    Inc( XPos, 8 );
  End;
End;

Procedure DisplayDocumentName( Var Name : String );
Begin
  FillChar( VGA^[ 8 * 24, 0 ], 320 * 8, StatusColor.B );
  WriteTextXY( 0, 8 * 24, 'File: '+Name, StatusColor.F );
End;
{-[ Copy VGA BIOS 8X8 Charset Into Gameboy Character Array ]----------}
Procedure CopyBIOS8X8CharSet( Buffer : pGBCArray );
Var TBIOS8X8   : pBIOSCArray;
    CharNumber : Integer;
    Y          : Integer;
    Tmp        : Byte;

Begin
  FillChar( Buffer^, SizeOf( Buffer^ ), 0 );
  { get address of 8X8 BIOS charset }
  ASM
    PUSH BP; MOV AH, $11; MOV AL, $30; MOV BH, $03; INT $10
    MOV AX, BP; POP BP; MOV Word( TBIOS8X8 ), AX;
    MOV Word( TBIOS8X8 ) + 2, ES
  END;
  BIOS8X8 := TBIOS8X8;
  For CharNumber := 0 to 255 do
  Begin
    For Y := 0 to 7 do
    Begin
      { next line will crash and burn in protected mode }
      Tmp := BIOS8X8^[ CharNumber, Y ];
      Buffer^[ CharNumber, Y * 2 ] := Tmp;
      Buffer^[ CharNumber, ( Y * 2 ) + 1 ] := Tmp;
    End;
  End;
End;

{-[ Draw Full Charset To Screen ]-------------------------------------}
Procedure DrawGameBoyCharset;
Var CX,   CY   : Integer;
    XPOS, YPOS : Integer;
    CharNum    : Integer;
    GBY        : Integer;
    Temp       : Word;
    TBL        : Word;
    Y          : Word;
Begin
  If CharWinState = W32x8
  Then
  Begin
    Box( 4 * 8 - 1, 11 * 4 - 1,
         32 * 8 + 2 , 8 * 8 + 2, StatusColor.F );
    For CY := 0 to 7 do
    Begin
      YPOS := ( CY + 4 ) * 8 + 4;
      For CX := 0 to 31 do
      Begin
        XPOS    := CX * 8;
        CharNum := CX + ( CY * 32 );
        GBY := 0;
        For Y := 0 to 7 do
        Begin
          Temp := CharBuffer^[ CharNum, GBY ];
          TBL  := CharBuffer^[ CharNum, GBY + 1 ];
          Move( GB2VGATbl[ TBL ]^[ Temp ] ,
                VGA^[ ( 1 * 8 ) + YPOS + Y,
                      ( 4 * 8 ) + XPOS ], 8 );
          Inc( GBY, 2 );
        End;
      End;
    End;
  End;
End;

Procedure UpdateGBCharsetActive( Index : Word );
Var CY   : Integer;
    XPos : Integer;
    YPos : Integer;
    Temp : Word;
    TBL  : Word;
Begin
  If CharWinState = W32x8
  Then
  Begin
    XPos := ( Index AND 31 ) * 8;
    YPos := ( ( Index SHR 5  ) * 8 ) + ( 8 * 4 ) + 4;
      For CY := 0 to 7 do
      Begin
        Temp := CharBuffer^[ Index, CY * 2 ];
        TBL  := CharBuffer^[ Index, ( CY * 2 ) + 1 ];
        Move( GB2VGATbl[ TBL ]^[ Temp ] ,
              VGA^[ ( 1 * 8 ) + YPOS + CY,
                    ( 4 * 8 ) + XPOS ], 8 );
      End;
  End;
End;

{-[ Set VGA DAC Regs Which Contain The 4 Game Boy Colors ]------------}
Procedure SetVGADAC2GBColors ( C1, C2, C3, C4 : Byte );
Begin
  Port[$3C8] := 16;
  Port[$3C9] := GBColors[C1].Red;
  Port[$3C9] := GBColors[C1].Green;
  Port[$3C9] := GBColors[C1].Blue;
  Port[$3C9] := GBColors[C2].Red;
  Port[$3C9] := GBColors[C2].Green;
  Port[$3C9] := GBColors[C2].Blue;
  Port[$3C9] := GBColors[C3].Red;
  Port[$3C9] := GBColors[C3].Green;
  Port[$3C9] := GBColors[C3].Blue;
  Port[$3C9] := GBColors[C4].Red;
  Port[$3C9] := GBColors[C4].Green;
  Port[$3C9] := GBColors[C4].Blue;
End;

{-[ Put Up A Window And Get Yes (TRUE) or No (FALSE) Response ]-------}
Function YesNoWindow( Title   : String;
                      Message : String ) : Boolean;
Var X,     Y     : Integer;
    XSize, YSize : Integer;
    Buffer       : PChar;
    TotalSize    : Word;
    YC           : Integer;
    Key          : Char;
Begin
    YesNoWindow := FALSE;

    If Length( Message ) > Length( Title )
    Then
      XSize := ( Length( Message ) * 8 ) + 10
    Else
      XSize := ( Length( Title ) * 8 ) + 10;

    YSize := 8 * 3 + 2;

    X := 160 - ( XSize DIV 2 );
    Y := 100 - ( YSize DIV 2 );
    TotalSize := XSize * YSize;

    GetMem( Buffer, TotalSize );
    For YC := 0 to YSize - 1 do
      Move( VGA^[ Y + YC, X ], Buffer[ YC * XSize ], XSize );

    Bar( X, Y, XSize, 9, StatusColor.B );
    Bar( X, Y + 9, XSize, YSize - 9, BackgroundColor.B );
    Box( X, Y, XSize, YSize, StatusColor.F );

    WriteTextXY( 160 - ( ( Length( Title ) * 8 ) DIV 2 ), Y + 1,
                 Title, StatusColor.F );
    WriteTextXY( 160 - ( ( Length( Message ) * 8 ) DIV 2 ),
                 Y + 1 + 8 + 4, Message, BackgroundColor.F );

    While KeyPressed do
      Readkey;

    Repeat
      Key := UpCase( Readkey );
      If Key = #0
      Then
        Readkey;
    Until ( Key = 'Y' ) OR ( Key = 'N' ) OR ( Key = #27 );
    YesNoWindow := Key = 'Y';

    For YC := 0 to YSize - 1 do
      Move( Buffer[ YC * XSize ], VGA^[ Y + YC, X ], XSize );
    FreeMem( Buffer, TotalSize );
End;

{-[ Convert Between Signed And Unsigned Char Formats ]----------------}
Procedure SwapCharBuffHalves;
Var TempCharBuff : pGBCArray;
Begin
  New( TempCharBuff );
  TempCharBuff^ := CharBuffer^;
  Move( TempCharBuff^[0], CharBuffer^[128], SizeOf( GBChar ) * 128 );
  Move( TempCharBuff^[128], CharBuffer^[0], SizeOf( GBChar ) * 128 );
  Dispose( TempCharBuff );
End;

{=[ UNDO BUFFER IMPLEMENTATION ]======================================}
{-[ Initialize Undo Buffer (Called *once* at start of program only )]-}
Procedure InitializeUndoBuffer;
Begin
  UndoBuffer := Nil;
End;

{-[ Save Undo ]-------------------------------------------------------}
Procedure SaveUndo( Buffer : pGBCArray );
Var TempUndo : UndoLink;
    Counter  : Integer;
Begin
  If UndoBuffer = Nil
  Then
  Begin
    New( UndoBuffer );
    UndoBuffer^.Next := Nil;
    UndoBuffer^.Data := Buffer^;
  End
  Else
  Begin
    TempUndo := UndoBuffer;
    Counter := 0;
    While TempUndo^.Next <> Nil do
    Begin
      TempUndo := TempUndo^.Next;
      Inc( Counter );
    End;
    New( TempUndo^.Next );
    TempUndo := TempUndo^.Next;
    TempUndo^.Next := Nil;
    TempUndo^.Data := Buffer^;

    If Counter > MaxUndo
    Then
    Begin
      TempUndo := UndoBuffer^.Next;
      Dispose( UndoBuffer );
      UndoBuffer := TempUndo;
    End;
  End;
End;

{-[ Undo Previous Edit ]----------------------------------------------}
Function RestoreUndo( Var Buffer : pGBCArray ) : Boolean;
Var TempUndo : UndoLink;
    Previous : UndoLink;
Begin
  If UndoBuffer = Nil
  Then
    RestoreUndo := FALSE
  Else
  Begin
    TempUndo := UndoBuffer;
    Previous := Nil;
    While TempUndo^.Next <> Nil do
    Begin
      Previous := TempUndo;
      TempUndo := TempUndo^.Next;
    End;
    Buffer^ := TempUndo^.Data;
    If Previous <> Nil
    Then
    Begin
      Dispose( TempUndo );
      Previous^.Next := Nil;
    End
    Else
    Begin
      Dispose( UndoBuffer );
      UndoBuffer := Nil;
    End;
    RestoreUndo := TRUE;
  End;
End;

{-[ Free Undo Buffer ]------------------------------------------------}
Procedure FreeUndoBuffer;
Var
  TempUndo : UndoLink;

Begin
  While UndoBuffer <> Nil do
  Begin
    TempUndo := UndoBuffer^.Next;
    Dispose( UndoBuffer );
    UndoBuffer := TempUndo;
  End;
End;

{-[ Make An Error Beep ]----------------------------------------------}
Procedure BeepError;
Begin
  Sound( 500 );
  Delay( 100 );
  NoSound;
End;

{-[ Get User Input String ]-------------------------------------------}
Procedure GetStringXY(     X, Y    : Integer;
                       Var S       : String;
                           MaxSize : Byte;
                           FColor,
                           BColor  : Byte);
Var
  Done    : Boolean;
  Key     : Char;
  CrsrPos : Integer;
  YC      : Integer;
Begin
  Box( X-1, Y-1,
       ( 8 * MaxSize ) + 2, 10, StatusColor.F );
  WriteTextXY( X, Y, S, BackgroundColor.F );

  { flush keyboard buffer }
  While Keypressed do
    Readkey;

  { get string loop }
  CrsrPos := 1;
  Done := FALSE;
  Repeat
    For YC := 0 to 7 do
      VGA^[ Y + YC, X + ( ( CrsrPos - 1 ) * 8 ) ] := Cursor1.Color;
    Key := ReadKey;
    Case Key of
      #0       : Case Ord(Readkey) of
                   Key_Delete : If( CrsrPos <= Length( S ) )
                                Then
                                Begin
                                  Delete( S, CrsrPos, 1 );
                                  Bar( X, Y,
                                       8 * MaxSize, 8,
                                       BackgroundColor.B );
                                  WriteTextXY( X, Y, S,
                                               BackgroundColor.F );
                                End
                                Else
                                  BeepError;
                   Key_LeftArrow  : If( CrsrPos > 1 )
                                    Then
                                    Begin
                                      Dec( CrsrPos, 1 );
                                      Bar( X, Y,
                                           8 * MaxSize, 8,
                                           BackgroundColor.B );
                                      WriteTextXY( X, Y, S,
                                                   BackgroundColor.F );
                                    End
                                    Else
                                      BeepError;
                   Key_RightArrow : If( CrsrPos <= Length( S ) )
                                    Then
                                    Begin
                                      Inc( CrsrPos, 1 );
                                      Bar( X, Y,
                                           8 * MaxSize, 8,
                                           BackgroundColor.B );
                                      WriteTextXY( X, Y, S,
                                                   BackgroundColor.F );
                                    End
                                    Else
                                      BeepError;
                 Else
                   BeepError;
                 End; { Case Readkey ( for extended keypress ) }
      '0'..'9',
      'A'..'Z',
      'a'..'z',
      '.', '\',
      ':', '~',
      '-', '+' : Begin
                   If( Length( S ) < MaxSize )
                   Then
                   Begin
                     {WriteTextXY( X + ( Length( S ) * 8 ), Y,
                                  Key, BackgroundColor.F );}
                     Insert( Key, S, CrsrPos );
                     Bar( X, Y,
                          8 * MaxSize, 8, BackgroundColor.B );
                     WriteTextXY( X, Y, S, BackgroundColor.F );
                     Inc( CrsrPos, 1 );
                   End
                   Else
                     BeepError;
                 End;
      #13      : Done := TRUE;
      #8       : If( ( Length( S ) > 0 ) AND ( CrsrPos > 1 ) )
                 Then
                 Begin
                   Delete( S, CrsrPos - 1, 1 );
                   Bar( X, Y,
                        8 * MaxSize, 8, BackgroundColor.B );
                   WriteTextXY( X, Y, S, BackgroundColor.F );
                   Dec( CrsrPos, 1 );
                 End
                 Else
                   BeepError;
      #27      : Begin
                   S := '';
                   Done := TRUE;
                 End;
    Else
      BeepError
    End; { Case Key }
  Until Done;
End;

{=[ File I/O Routines ]===============================================}
Procedure RequestFileName(     Title    : String;
                           Var FileName : String );
Var
  BkgBuffer : PChar;
  X, Y      : Integer;
  XSize,
  YSize     : Integer;
  TotalSize : Word;
  YC        : Integer;
Begin
  XSize := ( 30 * 8 ) + 8 + 2;
  YSize := ( 2 * 8 ) + 8 + 2;
  TotalSize := XSize * YSize;
  X := 160 - ( XSize DIV 2 );
  Y := 100 - ( YSize DIV 2 );
  GetMem( BkgBuffer, TotalSize );
  For YC := 0 to YSize - 1 do
    Move( VGA^[ Y + YC, X ], BkgBuffer[ YC * XSize ], XSize );

  Bar( X, Y, XSize, 9, StatusColor.B );
  Bar( X, Y + 9, XSize, YSize - 9, BackgroundColor.B );
  Box( X, Y, XSize, YSize, StatusColor.F );

  { Write title }
  WriteTextXY( 160 - ( Length( Title ) * 4 ),
               Y + 1, Title, StatusColor.F );
  WriteTextXY( X + 1 + 4, Y + 9 + 4, 'File:', BackgroundColor.F);

  GetStringXY( X + 1 + 4 + ( 5 * 8 ), Y + 9 + 4,
               FileName, 30 - 5,
               BackgroundColor.F, BackgroundColor.B );

  For YC := 0 to YSize - 1 do
      Move( BkgBuffer[ YC * XSize ], VGA^[ Y + YC, X ], XSize );
  FreeMem( BkgBuffer, TotalSize );
End;

{-[ Save Characterset to disk ]----------------------------------------}
Function SaveCharset : Boolean;

   Procedure SaveBinary( Var FName : String );
   Var FOut : File;
   Begin
     Assign( FOut, FName );
     Rewrite( FOut,1 );
     BlockWrite( FOut, CharSetFileID^, StrLen( CharSetFileID ) + 1);
     BlockWrite( FOut, CharBuffer^, SizeOf( CharBuffer^ ) );
     Close( FOut );
   End;

Var FileName : String;
Begin
  SaveCharset := FALSE;

  If CurrentFile = ''
  Then
    FileName := '.GBC'
  Else
    FileName := CurrentFile;

  RequestFileName( 'Enter File Name To Save', FileName );

  If Length( FileName ) > 0
  Then

    If FileExists( FileName )
    Then
    Begin
      If YesNoWindow( 'File Already Exists!',
                      'Replace existing file? (Y/N)' )
      Then
      Begin
        SaveBinary( FileName );
        CurrentFile := FileName;
        DisplayDocumentName( CurrentFile );
        SaveCharset := TRUE;
      End
    End
    Else
    Begin
      SaveBinary( FileName );
      CurrentFile := FileName;
      DisplayDocumentName( CurrentFile );
      SaveCharset := TRUE;
    End;
End;

{-[ Load Charset From Disk ]-------------------------------------------}
Function LoadCharset : Boolean;

  Function LoadBinary( Var FName : String ) : Boolean;
  Var FIn : File;
      FID : Array[ 0 .. 79 ] of Char;
      NumRead : Word;
  Begin
    LoadBinary := FALSE;
    FillChar( FID, 80, 0 );
    Assign( FIn, FName );
    Reset( FIn, 1 );
    BlockRead( FIn, FID, StrLen( CharSetFileID ) + 1 );
    If StrIComp( FID, CharSetFileID ) = 0
    Then
    Begin
      BlockRead( FIn, CharBuffer^, SizeOf( CharBuffer^ ), NumRead );
      If NumRead = SizeOf( CharBuffer^ )
      Then
        LoadBinary := TRUE;
    End;
    Close( FIn );
  End;

Var FileName : String;
Begin
  LoadCharset := False;
  FileName := '.GBC';
  RequestFileName( 'Enter File Name To Load', FileName );
  If( ( Length( FileName ) <> 0 ) AND
      ( FileName <> '.GBC' )      AND
      ( FileExists( FileName ) ) )
  Then
    If LoadBinary( FileName )
    Then
    Begin
      CurrentFile := FileName;
      DisplayDocumentName( CurrentFile );
      LoadCharset := TRUE;
    End;
End;

{-[ Export Characterset to asm file ]---------------------------------}
Function ExportCharset : Boolean;

   Procedure SaveText( Var FName : String );
   Var FOut      : Text;
       CharCount : Integer;
       YCount    : Integer;
   Begin
     Assign( FOut, FName );
     Rewrite( FOut );

     Writeln( FOut, '; Game Boy Character Set Made With GBChrEd by TechT' );
     Writeln( FOut, 'charset:' );
     For CharCount := 0 to 255 do
     Begin
       Write( FOut, '  .byte ' );
       For YCount := 0 to 6 do
         Write( FOut, CharBuffer^[ CharCount ][ YCount ] : 3, ',' );
       Writeln( FOut, CharBuffer^[ CharCount ][ 7 ] : 3 );
       Write( FOut, '  .byte ' );
       For YCount := 8 to 14 do
         Write( FOut, CharBuffer^[ CharCount ][ YCount ] : 3, ',' );
       Writeln( FOut, CharBuffer^[ CharCount ][ 15 ] : 3,
                      '   ; #',CharCount );
     End;

     Close( FOut );
   End;

Var FileName : String;
Begin
  ExportCharset := FALSE;

  FileName := '.INC';

  RequestFileName( 'Enter File Name To Export', FileName );

  If Length( FileName ) > 0
  Then

    If FileExists( FileName )
    Then
    Begin
      If YesNoWindow( 'File Already Exists!',
                      'Replace existing file? (Y/N)' )
      Then
      Begin
        SaveText( FileName );
        ExportCharset := TRUE;
      End
    End
    Else
    Begin
      SaveText( FileName );
      ExportCharset := TRUE;
    End;
End;

{=[ Main Program ]====================================================}
Var Counter : Integer;
Begin
  NoSound;
  CurrentFile := '';
  FillChar( CopyChar, SizeOf( CopyChar ), 0 );
  New( CharBuffer );
  InitializeUndoBuffer;
  CopyBIOS8x8CharSet( CharBuffer );
  BuildGB2VGATbl;
  VGA := Ptr( SegA000,0 );
  { go VGA 320x200x256 graphics mode }
  ASM MOV AX, $13; INT $10; END;
  SetVGADAC2GBColors( 0,1,2,3 );
  SetUpBackground;
  DrawGameBoyCharSet;
  Done := FALSE;
  DisplayStatus('Game Boy char set editor by TechT');
  {
  Box( 4 * 8 - 1, 11 * 4 - 1,
       32 * 8 + 1 , 8 * 8 + 1, StatusColor.F );
  Box( 12 * 8 - 1, 3 * 4 - 1,
       16 * 8 + 1, 16 * 8 + 1, StatusColor.F );
  }
  DrawPickCursor ( TRUE );
  Repeat
    DrawZoomCursor ( TRUE );
    Case UpCase(ReadKey) of
      #0  : Case Ord(Readkey) of
              Key_UpArrow    : If( ( MEM[ Seg0040:$17 ] AND 3 ) = 0 )
                               Then
                               Begin
                                 DrawZoomCursor( FALSE );
                                 Dec( Cursor1.Y );
                                 If Cursor1.Y < 0
                                 Then
                                 Begin
                                   Cursor1.Y := 7;
                                   Cursor1.Index :=
                                     ( Cursor1.Index - 32 ) AND 255;
                                   DrawPickCursor( TRUE );
                                   DrawZoomAll;
                                 End;
                               End
                               Else
                               Begin
                                 Cursor1.Index :=
                                   ( Cursor1.Index - 32 ) AND 255;
                                 DrawPickCursor( TRUE );
                                 DrawZoomAll;
                               End;

              Key_DownArrow  : If( ( MEM[ Seg0040:$17 ] AND 3 ) = 0 )
                               Then
                               Begin
                                 DrawZoomCursor( FALSE );
                                 Inc( Cursor1.Y );
                                 If Cursor1.Y > 7
                                 Then
                                 Begin
                                   Cursor1.Y := 0;
                                   Cursor1.Index :=
                                     ( Cursor1.Index + 32 ) AND 255;
                                   DrawPickCursor( TRUE );
                                   DrawZoomAll;
                                 End;
                               End
                               Else
                               Begin
                                 Cursor1.Index :=
                                   ( Cursor1.Index + 32 ) AND 255;
                                 DrawPickCursor( TRUE );
                                 DrawZoomAll;
                               End;
              Key_LeftArrow  : If( ( MEM[ Seg0040:$17 ] AND 3 ) = 0 )
                               Then
                               Begin
                                 DrawZoomCursor( FALSE );
                                 Dec( Cursor1.X );
                                 If Cursor1.X < 0
                                 Then
                                 Begin
                                   Cursor1.X := 7;
                                   Cursor1.Index :=
                                     ( Cursor1.Index - 1 ) AND 255;
                                   DrawPickCursor( TRUE );
                                   DrawZoomAll;
                                 End;
                               End
                               Else
                               Begin
                                 Cursor1.Index :=
                                   ( Cursor1.Index - 1 ) AND 255;
                                 DrawPickCursor( TRUE );
                                 DrawZoomAll;
                               End;
              Key_RightArrow : If( ( MEM[ Seg0040:$17 ] AND 3 ) = 0 )
                               Then
                               Begin
                                 DrawZoomCursor( FALSE );
                                 Inc( Cursor1.X );
                                 If Cursor1.X > 7
                                 Then
                                 Begin
                                   Cursor1.X := 0;
                                   Cursor1.Index :=
                                     ( Cursor1.Index + 1 ) AND 255;
                                   DrawPickCursor( TRUE );
                                   DrawZoomAll;
                                 End;
                               End
                               Else
                               Begin
                                 Cursor1.Index :=
                                   ( Cursor1.Index + 1 ) AND 255;
                                 DrawPickCursor( TRUE );
                                 DrawZoomAll;
                               End;

            {--- clear current character ---}
              Key_Delete : Begin
                             SaveUndo( CharBuffer );
                             For Counter := 0 to 15 do
                               CharBuffer^[ Cursor1.Index ][ Counter ] := 0;
                             UpdateZoomActive( Cursor1.Index );
                             DrawPickCursor( FALSE );
                             UpdateGBCharsetActive( Cursor1.Index );
                             DrawPickCursor( TRUE );
                             WorkChanged := TRUE;
                             DisplayStatus( 'Cleared Char Under Cursor' );
                           End;

            End;
      #27 : Begin
              If WorkChanged
              Then
              Begin
                If YesNoWindow( 'Work Not Saved!',
                              'Do you really want to exit? (Y/N)' )
                Then
                  Done := TRUE
              End
              Else
                Done := TRUE;
            End;

    {--- edit pixel under cursor ---}
      '1' : Begin
              SaveUndo( CharBuffer );
              CharBuffer^[ Cursor1.Index ][ Cursor1.Y * 2 ] :=
                CharBuffer^[ Cursor1.Index ][ Cursor1.Y * 2 ] AND
                NOT( 128 SHR Cursor1.X );
              CharBuffer^[ Cursor1.Index ][ ( Cursor1.Y * 2 ) + 1 ] :=
                CharBuffer^[ Cursor1.Index ][ ( Cursor1.Y * 2 ) + 1  ]
                AND NOT( 128 SHR Cursor1.X );
              UpdateZoomActive( Cursor1.Index );
              DrawPickCursor( FALSE );
              UpdateGBCharsetActive( Cursor1.Index );
              DrawPickCursor( TRUE );
              WorkChanged := TRUE;
            End;
      '2' : Begin
              SaveUndo( CharBuffer );
              CharBuffer^[ Cursor1.Index ][ Cursor1.Y * 2 ] :=
                CharBuffer^[ Cursor1.Index ][ Cursor1.Y * 2 ] OR
                ( 128 SHR Cursor1.X );
              CharBuffer^[ Cursor1.Index ][ ( Cursor1.Y * 2 ) + 1 ] :=
                CharBuffer^[ Cursor1.Index ][ ( Cursor1.Y * 2 ) + 1  ]
                AND NOT( 128 SHR Cursor1.X );
              UpdateZoomActive( Cursor1.Index );
              DrawPickCursor( FALSE );
              UpdateGBCharsetActive( Cursor1.Index );
              DrawPickCursor( TRUE );
              WorkChanged := TRUE;
            End;
      '3' : Begin
              SaveUndo( CharBuffer );
              CharBuffer^[ Cursor1.Index ][ Cursor1.Y * 2 ] :=
                CharBuffer^[ Cursor1.Index ][ Cursor1.Y * 2 ] AND
                NOT( 128 SHR Cursor1.X );
              CharBuffer^[ Cursor1.Index ][ ( Cursor1.Y * 2 ) + 1 ] :=
                CharBuffer^[ Cursor1.Index ][ ( Cursor1.Y * 2 ) + 1  ]
                OR ( 128 SHR Cursor1.X );
              UpdateZoomActive( Cursor1.Index );
              DrawPickCursor( FALSE );
              UpdateGBCharsetActive( Cursor1.Index );
              DrawPickCursor( TRUE );
              WorkChanged := TRUE;
            End;
      '4' : Begin
              SaveUndo( CharBuffer );
              CharBuffer^[ Cursor1.Index ][ Cursor1.Y * 2 ] :=
                CharBuffer^[ Cursor1.Index ][ Cursor1.Y * 2 ] OR
                ( 128 SHR Cursor1.X );
              CharBuffer^[ Cursor1.Index ][ ( Cursor1.Y * 2 ) + 1 ] :=
                CharBuffer^[ Cursor1.Index ][ ( Cursor1.Y * 2 ) + 1  ]
                OR ( 128 SHR Cursor1.X );
              UpdateZoomActive( Cursor1.Index );
              DrawPickCursor( FALSE );
              UpdateGBCharsetActive( Cursor1.Index );
              DrawPickCursor( TRUE );
              WorkChanged := TRUE;
            End;
      #32,
      '5' : Begin
              SaveUndo( CharBuffer );
              CharBuffer^[ Cursor1.Index ][ Cursor1.Y * 2 ] :=
                CharBuffer^[ Cursor1.Index ][ Cursor1.Y * 2 ] XOR
                ( 128 SHR Cursor1.X );
              CharBuffer^[ Cursor1.Index ][ ( Cursor1.Y * 2 ) + 1 ] :=
                CharBuffer^[ Cursor1.Index ][ ( Cursor1.Y * 2 ) + 1  ]
                XOR ( 128 SHR Cursor1.X );
              UpdateZoomActive( Cursor1.Index );
              DrawPickCursor( FALSE );
              UpdateGBCharsetActive( Cursor1.Index );
              DrawPickCursor( TRUE );
              WorkChanged := TRUE;
            End;

    {--- negate character under cursor ---}
      'N' : Begin
              SaveUndo( CharBuffer );
              For Counter := 0 to 15 do
                CharBuffer^[ Cursor1.Index ][ Counter ] :=
                  CharBuffer^[ Cursor1.Index ][ Counter ] XOR $FF;
              UpdateZoomActive( Cursor1.Index );
              DrawPickCursor( FALSE );
              UpdateGBCharsetActive( Cursor1.Index );
              DrawPickCursor( TRUE );
              WorkChanged := TRUE;
              DisplayStatus( 'Negated Char Under Cursor' );
            End;

    {--- convert to/from signed and unsigned charsets ---}
      '-' : Begin
              SaveUndo( CharBuffer );
              SwapCharBuffHalves;
              Cursor1.Index := ( Cursor1.Index + 128 ) AND 255;
              DrawZoomAll;
              DrawPickCursor( FALSE );
              DrawGameBoyCharset;
              DrawPickCursor( TRUE );
              WorkChanged := TRUE;
              DisplayStatus( 'Changed Sign' );
            End;

    {--- copy undo buffer over current character buffer ---}
      #08 : Begin
              If NOT( RestoreUndo( CharBuffer ) )
              Then
                DisplayStatus( 'Can''t Undo' )
              Else
              Begin
                DrawZoomAll;
                DrawPickCursor( FALSE );
                DrawGameBoyCharSet;
                DrawPickCursor( TRUE );
                DisplayStatus( 'Undid Changes' );
                WorkChanged := TRUE;
              End;
            End;

    {--- save charset to disk ---}
      'S' : Begin
              DisplayStatus( '"Save" Requested...' );
              If SaveCharset
              Then
              Begin
                WorkChanged := FALSE;
                FreeUndoBuffer;
                DisplayStatus( '"Save" Successful' );
              End
              Else
                DisplayStatus( '"Save" Failed Or Aborted' );
            End;
    {--- load charset from disk ---}
      'L' : Begin
              DisplayStatus( '"Load" Requested...' );
              If LoadCharset
              Then
              Begin
                WorkChanged := FALSE;
                FreeUndoBuffer;
                DisplayStatus( '"Load" Successful' );
              End
              Else
                DisplayStatus( '"Load" Failed Or Aborted' );
              DrawZoomAll;
              DrawPickCursor( FALSE );
              DrawGameBoyCharSet;
              DrawPickCursor( TRUE );
            End;
    {--- export charset to assembler file ---}
      'X' : Begin
              DisplayStatus( '"Export" Requested...' );
              If ExportCharset
              Then
                DisplayStatus( '"Export" Successful' )
              Else
                DisplayStatus( '"Export" Failed Or Aborted' );
            End;

    {--- copy character under cursor ---}
      'C' : Begin
              CopyChar := CharBuffer^[ Cursor1.Index ];
              DisplayStatus( 'Copied Char Under Cursor' );
            End;
    {--- paste buffer to character under cursor ---}
      'P' : Begin
              {If( GBChar( CharBuffer^[ Cursor1.Index ] ) <>
                  GBChar( CopyChar ) )
              Then
              Begin}
                SaveUndo( CharBuffer );
                CharBuffer^[ Cursor1.Index ] := CopyChar;
                UpdateZoomActive( Cursor1.Index );
                DrawPickCursor( FALSE );
                UpdateGBCharsetActive( Cursor1.Index );
                DrawPickCursor( TRUE );
                WorkChanged := TRUE;
                DisplayStatus( 'Pasted Over Char Under Cursor' );
              {End
              Else
                DisplayStatus( 'Pasting Here Would Be Redundant' );}
            End;
    End;

  Until Done;

  TextMode( C80 );

  FreeUndoBuffer;
  Dispose( CharBuffer );

  Writeln( 'GBChrEd V'+Version+' by TechT' );
  Writeln( 'This program is PUBLIC DOMAIN' );
End.